function [x,f,info] = alm(fg,cxg,x0,csize,varargin)

p = inputParser;
p.addParamValue('maxiter',100,@(x) isnumeric(x) && x >= 0);
p.addParamValue('glaxtol',1e-6, @(x) isnumeric(x) && x >= 0);
p.addParamValue('ctol',1e-6, @(x) isnumeric(x) && x >= 0);
p.addParamValue('cpow',0.1, @(x) isnumeric(x) && x > 0 && x < 1);
p.addParamValue('glaxpowa', 0.1, @(x) isnumeric(x) && x > 0 && x < 1);
p.addParamValue('glaxpowb', 0.9, @(x) isnumeric(x) && x > 0);
p.addParamValue('quiet',0);
p.addParamValue('rho', 10, @(x) isnumeric(x) && x > 0);
p.addParamValue('tau', 100, @(x) isnumeric(x) && x > 0);

p.parse(varargin{:});
opts = p.Results;

x = x0;
n = numel(x);
m = csize;

f = Inf;
g = zeros(n,1);
y = zeros(m,1);
cx = zeros(m,1);
rhok = opts.rho;
glaxtolk = 1.0/rhok;
ctolk = 1.0/(rhok^opts.cpow);

%if ~opts.quiet
%    fprintf('  %6s  %9s %9s %9s\n', 'iter', ...
%        'fval', 'normg', 'fdiff');
%end

for iter=1:opts.maxiter
    
    xp = x;
    yp = y;

    %pbout = ncg(@(xx) flax(xx, fg, cxg, y, rhok), x);
    %x = pbout.X;
    %g = pbout.G;
    %dispf = pbout.F;
    
    options = optimset('GradObj','on');
    [x,dispf,junk2,junk3,g] = fminunc(@(xx) flax(xx, fg, cxg, y, rhok), x, options);
    
    cx = cxg(x);
    
    xnorm = norm(x,'inf');
    gnorm = norm(g,'inf');
    cnorm = norm(cx,'inf');
    
    fprintf('Iter: %d, fx: %f, xnorm: %f, gnorm: %f, cnorm: %f, rhok: %d\n',iter, dispf, xnorm, gnorm, cnorm, rhok);
    
    if ( cnorm <= max([opts.ctol, ctolk]) )
        if ( cnorm <= opts.ctol && gnorm <= opts.glaxtol )
            f = fg(x);
            info.exitflag = 0;
            break;
        end
        
        y = y - rhok * cx;
        ctolk = ctolk / (rhok^(1-opts.cpow));
        glaxtolk = glaxtolk / (rhok^(opts.glaxpowb));
        
    else
        x = xp;
        y = yp;
        rhok = rhok * opts.tau;
        ctolk = 1.0 / (rhok^opts.cpow);
        glaxtolk = glaxtolk / (rhok^(opts.glaxpowa));
        
    end
    
end

if (iter==opts.maxiter)
    f = fg(x);
    info.exitflag=3;
end

if ( gnorm > opts.glaxtol || cnorm > opts.ctol )
    warning('Did not converge');
end

