#include <stdio.h>
#include <stdlib.h>
#include <lbfgs.h>
#include <math.h>
#include "tbb/task_scheduler_init.h"
#include "tbb/tick_count.h"

using namespace tbb;
using namespace std;


static lbfgsfloatval_t xij_step(
    void *instance,
    const lbfgsfloatval_t *x,
    lbfgsfloatval_t *g,
    const int n,
    const lbfgsfloatval_t step
    )
{

  // In instance, store [zi,zj,yij,yji,rho]; in x, store [xi,xj]

    int i;
    lbfgsfloatval_t tmp;

    lbfgsfloatval_t *zyarray=(lbfgsfloatval_t *)instance;

    lbfgsfloatval_t fx = 0.0;

    lbfgsfloatval_t xij=0.0;
    lbfgsfloatval_t yxzi=0.0;
    lbfgsfloatval_t xzi=0.0;    
    lbfgsfloatval_t yxzj=0.0;
    lbfgsfloatval_t xzj=0.0; 
    
    lbfgsfloatval_t rho=zyarray[2*n];
    
    // assert(i%2==0);
    int nn=n/2;

    for (i=0;i<nn;i++)
      {
	tmp=x[i]-x[nn+i];
	xij+=tmp*tmp;
      }

    for (i=0;i<nn;i++)
      {
	tmp=x[i]-zyarray[i];
	yxzi+=zyarray[i+nn*2]*tmp;
	xzi+=tmp*tmp;
      }

    for (i=0;i<nn;i++)
      {
	tmp=x[i+nn]-zyarray[i+nn];
	yxzj+=zyarray[i+nn*3]*tmp;
	xzj+=tmp*tmp;
      }
    
    for (i = 0;i < nn;i++) {
      tmp=x[i]-x[nn+i];
      g[i]=-2.0/xij/xij*tmp+zyarray[i+nn*2]+rho*(x[i]-zyarray[i]);
      g[nn+i]=2.0/xij/xij*tmp+zyarray[i+nn*3]+rho*(x[i+nn]-zyarray[i+nn]);	
    }
    fx=1.0/xij+yxzi+0.5*rho*xzi+yxzj+0.5*rho*xzj;

    return fx;
}


static int progress(
    void *instance,
    const lbfgsfloatval_t *x,
    const lbfgsfloatval_t *g,
    const lbfgsfloatval_t fx,
    const lbfgsfloatval_t xnorm,
    const lbfgsfloatval_t gnorm,
    const lbfgsfloatval_t step,
    int n,
    int k,
    int ls
    )
{
  //printf("Iteration %d:\n", k);
  //  printf("  fx = %f, x[0] = %f, x[1] = %f\n", fx, x[0], x[1]);
  //  printf("  xnorm = %f, gnorm = %f, step = %f\n", xnorm, gnorm, step);
  //  printf("\n");
    return 0;
}


int main(int argc, char *argv[])
{  

  task_scheduler_init init;

  int npts=3;
  if (argc>=2)
    npts=atoi(argv[1]);

  int dim=3;
  if (argc>=3)
    dim=atoi(argv[2]);

  lbfgsfloatval_t rho=10;
  if (argc>=4)
    rho=atof(argv[3]);

  int maxiter=10000;
  if (argc>=5)
    maxiter=atoi(argv[4]);

  lbfgsfloatval_t tol=1e-8;
  if (argc>=6)
    tol=atof(argv[5]);

  int i, j, d, n,iter,ret;

  printf("npts=%d\n",npts);
  printf("dim=%d\n",dim);
  printf("rho=%lf\n",rho);
  printf("maxiter=%d\n",maxiter);
  printf("tol=%e\n",tol);

  char filename[20];
  sprintf(filename,"ADMM_%d_%d.txt",npts,dim);

  FILE *fp=fopen(filename,"w");
  fprintf(fp,"npts=%d\n",npts);
  fprintf(fp,"dim=%d\n",dim);
  fprintf(fp,"rho=%lf\n",rho);
  fprintf(fp,"maxiter=%d\n",maxiter);
  fprintf(fp,"tol=%e\n",tol);
  fprintf(fp,"algo=ADMM\n");
  

    // allocate xij zi yij
  lbfgsfloatval_t *x=lbfgs_malloc(dim*npts*npts);
  lbfgsfloatval_t *y=lbfgs_malloc(dim*npts*npts);
  lbfgsfloatval_t *z=lbfgs_malloc(dim*npts);
    
  srand(1);
  lbfgsfloatval_t *initz=lbfgs_malloc(dim);
  lbfgsfloatval_t *tmpx=lbfgs_malloc(dim);
  lbfgsfloatval_t sumz,sumx,tmpy,diff;
  
  // initialize z
  for (i=0;i<npts;i++)
    {
      sumz=0;
      for (d=0;d<dim;d++)
	{
	  initz[d]=rand()%1000-500;
	  sumz+=initz[d]*initz[d];
	}
      sumz=sqrt(sumz);
      for (d=0;d<dim;d++)
	  z[dim*i+d]=initz[d]/sumz;      
    }

  // initialize x
  for (i=0;i<npts;i++)
    {
      for (j=0;j<npts;j++)
	{
	  for (d=0;d<dim;d++)
	    {
	      x[dim*npts*i+dim*j+d]=z[i*dim+d];
	    }
	}
    }

  // initialize y
  for (i=0;i<npts;i++)
    {
      for (j=0;j<npts;j++)
	{
	  for (d=0;d<dim;d++)
	    {
	      y[dim*npts*i+dim*j+d]=0.0;
	    }
	}
    }


  /************* Begin ADMM ***************/

  int converg_flag=-1;
  lbfgsfloatval_t fx;
  lbfgs_parameter_t param;
  lbfgsfloatval_t *x_array=lbfgs_malloc(dim*2);
  lbfgsfloatval_t *zy_array = lbfgs_malloc(dim*4+1);

  tick_count serial_t0 = tick_count::now();
  for (iter=0;iter<maxiter;iter++)
    {
      printf("********* Outer Loop Iteration: %d\n",iter);
      // optimize xij
      for ( n=0;n<npts*(npts-1)/2;n++)
	{
	  i=(int)((1+sqrt(1+8.0*n))/2);
	  j=n-(i-1)*i/2;

	      for (d=0;d<dim;d++)
		{
		  x_array[d]=x[dim*npts*i+dim*j+d];  // xij
		  x_array[d+dim]=x[dim*npts*j+dim*i+d];	// xji	
		}
	      for (d=0;d<dim;d++)
		{
		  zy_array[d]=z[dim*i+d];   // zi
		  zy_array[d+dim]=z[dim*j+d];  // zj
		  zy_array[d+dim*2]=y[dim*npts*i+dim*j+d];  // yij
		  zy_array[d+dim*3]=y[dim*npts*j+dim*i+d];  // yji
		}
	      zy_array[dim*4]=rho;
		
	      lbfgs_parameter_init(&param);
	      ret = lbfgs(dim*2, x_array, &fx, xij_step, progress, zy_array, &param);
	      /* Report the result. */
	      //printf("L-BFGS optimization terminated with status code = %d\n", ret);
	      //printf("  fx = %f\n", fx);

	      // store back to x
	      for (d=0;d<dim;d++)
		{
		  x[dim*npts*i+dim*j+d]=x_array[d];  // xij
		  x[dim*npts*j+dim*i+d]=x_array[dim+d];	// xji	
		}
	}
      
      // optimize xii
      for (i=0;i<npts;i++)
	{
	  sumx=0.0;
	  for (d=0;d<dim;d++)
	    {
	      tmpx[d]=z[dim*i+d]-y[dim*npts*i+dim*i+d]/rho;
	      sumx+=tmpx[d]*tmpx[d];
	    }
	  sumx=sqrt(sumx);
	  for(d=0;d<dim;d++)
	    {
	      x[dim*npts*i+dim*i+d]=tmpx[d]/sumx;  // project to norm ball
	    }
	}

      // optimize zi
      for (i=0;i<npts;i++)
	{
	  for (d=0;d<dim;d++)
	    {
	      z[i*dim+d]=0;
	      for (j=0;j<npts;j++)
		{
		  z[i*dim+d]+=(x[dim*npts*i+dim*j+d]+y[dim*npts*i+dim*j+d]/rho)/npts;
		}
	    }
	}
	
      // optimize yij
      diff=0;
      for (i=0;i<npts;i++)
	{
	  for (j=0;j<npts;j++)
	    {
	      for (d=0;d<dim;d++)
		{
		  tmpy=(x[dim*npts*i+dim*j+d]-z[i*dim+d]);
		  y[dim*npts*i+dim*j+d]+=tmpy*rho;
		  diff+=tmpy*tmpy;
		}
	    }
	}
      
      printf("diff=%e\n",diff);
      if (diff<tol)
	{
	  converg_flag=1;
	  break;
	}

    }

  tick_count serial_t1 = tick_count::now();

  printf("Converge Flag= %d \n Time Spent = %f \n",converg_flag,(serial_t1 - serial_t0).seconds());
  fprintf(fp,"Converge Flag= %d \n Time Spent = %f \n",converg_flag,(serial_t1 - serial_t0).seconds());


  for (i=0;i<npts;i++)
    {
      sumz=0.0;
      for (d=0;d<dim;d++)
	{
	  fprintf(fp,"z[%d][%d]=%lf, ",i,d,z[i*dim+d]);
	  sumz+=z[i*dim+d]*z[i*dim+d];
	}
      fprintf(fp,"|z[%d]|=%lf \n",i,sumz);
    }

  fx=0.0;
  for (i=0;i<npts;i++)
    {
      for (j=i+1;j<npts;j++)
	{
	  sumz=0.0;
	  for (d=0;d<dim;d++)
	    {
	      sumz+=(z[i*dim+d]-z[j*dim+d])*(z[i*dim+d]-z[j*dim+d]);      
	    }
	  fx+=1.0/sumz;
	}
    }

  fprintf(fp,"fx=%lf\n",fx);  
  printf("fx=%lf\n",fx);
  
  fclose(fp);

  lbfgs_free(x_array);
  lbfgs_free(zy_array);
  lbfgs_free(x);
  lbfgs_free(y);
  lbfgs_free(z);
  lbfgs_free(initz);
  lbfgs_free(tmpx);

  return 0;
}
