#include <stdio.h>
#include <stdlib.h>
#include <lbfgs.h>
#include <math.h>
#include "tbb/task_scheduler_init.h"
#include "tbb/parallel_for.h"
#include "tbb/parallel_reduce.h"
#include "tbb/blocked_range.h"
#include "tbb/tick_count.h"

using namespace tbb;
using namespace std;

#define min2(a, b)      ((a) <= (b) ? (a) : (b))
#define max2(a, b)      ((a) >= (b) ? (a) : (b))

// global variables
lbfgsfloatval_t *x;
lbfgsfloatval_t *y;
lbfgsfloatval_t *z;
static int npts;
static int dim;
static lbfgsfloatval_t rho;
static lbfgsfloatval_t all_cnorm,all_gnorm;

static lbfgsfloatval_t xij_funeval(
    void *instance,
    const lbfgsfloatval_t *x,
    lbfgsfloatval_t *g,
    const int n,
    const lbfgsfloatval_t step
    )
{

  // In instance, store [zi,zj,yij,yji,rho]; in x, store [xi,xj]

    int i;
    lbfgsfloatval_t tmp;

    lbfgsfloatval_t *zyarray = (lbfgsfloatval_t *)instance;

    lbfgsfloatval_t fx = 0.0;

    lbfgsfloatval_t xij=0.0;
    lbfgsfloatval_t yxzi=0.0;
    lbfgsfloatval_t xzi=0.0;    
    lbfgsfloatval_t yxzj=0.0;
    lbfgsfloatval_t xzj=0.0; 
    
   
    // assert(i%2==0);
    int nn=n/2;

    for (i=0;i<nn;i++)
      {
	tmp=x[i]-x[nn+i];
	xij+=tmp*tmp;
      }

    for (i=0;i<nn;i++)
      {
	tmp=x[i]-zyarray[i];
	yxzi+=zyarray[i+nn*2]*tmp;
	xzi+=tmp*tmp;
      }

    for (i=0;i<nn;i++)
      {
	tmp=x[i+nn]-zyarray[i+nn];
	yxzj+=zyarray[i+nn*3]*tmp;
	xzj+=tmp*tmp;
      }
    
    for (i = 0;i < nn;i++) {
      tmp=x[i]-x[nn+i];
      g[i]=-2.0/xij/xij*tmp/(npts*(npts-1)/2)+zyarray[i+nn*2]+rho*(x[i]-zyarray[i]);
      g[nn+i]=2.0/xij/xij*tmp/(npts*(npts-1)/2)+zyarray[i+nn*3]+rho*(x[i+nn]-zyarray[i+nn]);	
    }
    fx=1.0/xij/(npts*(npts-1)/2)+yxzi+0.5*rho*xzi+yxzj+0.5*rho*xzj;

    return fx;
}


static int progress(
    void *instance,
    const lbfgsfloatval_t *x,
    const lbfgsfloatval_t *g,
    const lbfgsfloatval_t fx,
    const lbfgsfloatval_t xnorm,
    const lbfgsfloatval_t gnorm,
    const lbfgsfloatval_t step,
    int n,
    int k,
    int ls
    )
{
  //printf("Iteration %d:\n", k);
  //  printf("  fx = %f, x[0] = %f, x[1] = %f\n", fx, x[0], x[1]);
  //  printf("  xnorm = %f, gnorm = %f, step = %f\n", xnorm, gnorm, step);
  //  printf("\n");
    return 0;
}



class xij_step {
public:
  lbfgsfloatval_t gnorm;

  void operator() ( const blocked_range<size_t>& r ) const {

    int i,j,n,d;
    lbfgs_parameter_t param;
    param.epsilon = (1e-5)/(npts*(npts-1)/2);

    lbfgsfloatval_t fx;    
    lbfgsfloatval_t *x_array=new lbfgsfloatval_t[dim*2];
    lbfgsfloatval_t *zy_array=new lbfgsfloatval_t[dim*4];	


    for (n=r.begin();n<r.end();n++)
      {

	i=(int)((1+sqrt(1+8.0*n))/2);
	j=n-(i-1)*i/2;
	
	for (d=0;d<dim;d++)
	  {
	    x_array[d]=x[dim*npts*i+dim*j+d];  // xij
	    x_array[d+dim]=x[dim*npts*j+dim*i+d];	// xji	
	  }
	for (d=0;d<dim;d++)
	  {
	    zy_array[d]=z[dim*i+d];   // zi
	    zy_array[d+dim]=z[dim*j+d];  // zj
	    zy_array[d+dim*2]=y[dim*npts*i+dim*j+d];  // yij
	    zy_array[d+dim*3]=y[dim*npts*j+dim*i+d];  // yji
	  }
		
	lbfgs_parameter_init(&param);
	//int ret=
	  lbfgs(dim*2, x_array, &fx, xij_funeval, progress, zy_array, &param);
	//if (n==0)
	//  printf("ret=%d\n",ret);
  
	for (d=0;d<dim;d++)
	  {
	    x[dim*npts*i+dim*j+d]=x_array[d];  // xij
	    x[dim*npts*j+dim*i+d]=x_array[dim+d];	// xji	
	  }
      }
    delete []x_array;
    delete []zy_array;	
    
  }

  xij_step()  {}
};




class xii_step {
public:

  void operator() ( const blocked_range<size_t>& r ) const {

    lbfgsfloatval_t *tmpx=new lbfgsfloatval_t[dim];

    lbfgsfloatval_t sumx;
    
    for (size_t i=r.begin();i<r.end();i++)
      {	

	sumx=0.0;
	for (size_t d=0;d<dim;d++)
	  {
	    tmpx[d]=z[dim*i+d]-y[dim*npts*i+dim*i+d]/rho;
	    sumx+=tmpx[d]*tmpx[d];
	  }
	sumx=sqrt(sumx);
	for(size_t d=0;d<dim;d++)
	  {
	    x[dim*npts*i+dim*i+d]=tmpx[d]/sumx;  // project to norm ball
	  }

      }
    delete []tmpx;

  }

  xii_step()  {}
};



class zi_step {
public:
  lbfgsfloatval_t gnorm;

  void operator() ( const blocked_range<size_t>& r ) {
    lbfgsfloatval_t tmp;
    for (size_t i=r.begin();i<r.end();i++)
      {
	for (size_t d=0;d<dim;d++)
	  {
	    tmp=0;
	    for (size_t j=0;j<npts;j++)
	      {
		tmp+=(x[dim*npts*i+dim*j+d]+y[dim*npts*i+dim*j+d]/rho)/npts;		
	      }
	    gnorm+=(z[i*dim+d]-tmp)*(z[i*dim+d]-tmp)*rho*rho;
	    z[i*dim+d]=tmp;
	  }
      }
  }


  void join (zi_step &zi) {gnorm +=zi.gnorm;}
  zi_step() : gnorm(0) {}
  zi_step(zi_step &zi,split) {gnorm=0;}

};



class yij_step {
public:
  lbfgsfloatval_t cnorm;

  void operator() ( const blocked_range<size_t>& r ) {
    lbfgsfloatval_t tmpy;
    for (size_t i=r.begin();i<r.end();i++)
      {
	for (size_t j=0;j<npts;j++)
	  {
	    for (size_t d=0;d<dim;d++)
	      {
		tmpy=(x[dim*npts*i+dim*j+d]-z[i*dim+d]);
		y[dim*npts*i+dim*j+d]+=tmpy*rho;
		cnorm+=tmpy*tmpy;
	      }
	  }
      }
  }
  void join (yij_step &yij) {cnorm +=yij.cnorm;}
  yij_step() : cnorm(0) {}
  yij_step(yij_step &yij,split) {cnorm=0;}

};


static void parallel_xyz()
{
  // optimize xij
  xij_step xij;
  parallel_for(blocked_range<size_t>(0, npts*(npts-1)/2),xij,auto_partitioner());

  // optimize xii
  xii_step xii;
  parallel_for(blocked_range<size_t>(0, npts),xii,auto_partitioner());
	
  // optimize yij
  yij_step yij;
  parallel_reduce(blocked_range<size_t>(0, npts),yij,auto_partitioner());
  all_cnorm=sqrt(yij.cnorm);

  // optimize zi
  zi_step zi; 
  parallel_reduce(blocked_range<size_t>(0, npts),zi,auto_partitioner());
  all_gnorm=sqrt(zi.gnorm);

}


int main(int argc, char *argv[])
{  

  npts=3;
  if (argc>=2)
    npts=atoi(argv[1]);

  dim=3;
  if (argc>=3)
    dim=atoi(argv[2]);

  int rho_scale=1;
  rho=10.0;

  if (argc>=4)
    {
      rho_scale=atoi(argv[3]);
      rho=pow(10.0,rho_scale);
    }

  int maxiter=1000;
  if (argc>=5)
    maxiter=atoi(argv[4]);

  lbfgsfloatval_t tao=1.0;
  if (argc>=6)
    tao=atof(argv[5]);

  int nThreads=1;
  if (argc>=7)
    nThreads=atoi(argv[6]);

  lbfgsfloatval_t alpha=0.5;
  if (argc>=8)
    alpha=atof(argv[7]);

  lbfgsfloatval_t beta=0.01;
  if (argc>=9)
    beta=atof(argv[8]);



  task_scheduler_init init(nThreads);

  lbfgsfloatval_t tol=1e-6;
  int i, j, d, r,iter,ret;

  printf("npts=%d\n",npts);
  printf("dim=%d\n",dim);
  printf("rho=%lf\n",rho);
  printf("maxiter=%d\n",maxiter);
  printf("tol=%e\n",tol);
  printf("tao=%lf\n",tao);
  printf("alpha=%lf\n",alpha);
  printf("beta=%lf\n",beta);



  char filename1[80];
  char filename2[80];

  sprintf(filename1,"result_tbbadmm_%d_%d_%d_%.2lf_%d_%.1lf_%.2lf.txt",npts,dim,rho_scale,tao,nThreads,alpha,beta);
  sprintf(filename2,"x_tbbadmm_%d_%d_%d_%.2lf_%d_%.1lf_%.2lf.txt",npts,dim,rho_scale,tao,nThreads,alpha,beta);

  FILE *fp1=fopen(filename1,"w");
  FILE *fp2=fopen(filename2,"w");

    // allocate xij zi yij
  x=new lbfgsfloatval_t[dim*npts*npts];
  y=new lbfgsfloatval_t[dim*npts*npts];
  z=new lbfgsfloatval_t[dim*npts];

    
  srand(1);
  lbfgsfloatval_t *initz=new lbfgsfloatval_t[dim];
  lbfgsfloatval_t sumz,sumx,tmpy;
  
  // initialize z
  for (i=0;i<npts;i++)
    {
      sumz=0;
      for (d=0;d<dim;d++)
	{
	  initz[d]=rand()%1000-500;
	  sumz+=initz[d]*initz[d];
	}
      sumz=sqrt(sumz);
      for (d=0;d<dim;d++)
	  z[dim*i+d]=initz[d]/sumz;      
    }

  // initialize x
  for (i=0;i<npts;i++)
    {
      for (j=0;j<npts;j++)
	{
	  for (d=0;d<dim;d++)
	    {
	      x[dim*npts*i+dim*j+d]=z[i*dim+d];
	    }
	}
    }

  // initialize y
  for (i=0;i<npts;i++)
    {
      for (j=0;j<npts;j++)
	{
	  for (d=0;d<dim;d++)
	    {
	      y[dim*npts*i+dim*j+d]=0.0;
	    }
	}
    }


  /************* Begin ADMM ***************/

  int converg_flag=-1;

  // parameters for rho schedule
  lbfgsfloatval_t tolk=1.0;

  tick_count parallel_t0 = tick_count::now();
  for (iter=0;iter<maxiter;iter++)
    {
      
      fprintf(fp1,"%d ",iter);
   
      parallel_xyz();
      
      //rho schedule
      if (all_cnorm<tolk && all_gnorm<tolk)
	{
	  tolk=tolk/pow(rho,beta);
	  tolk=max2(tolk,tol);
	}
      else if (all_cnorm>tolk && all_gnorm<tolk)
	{
	  tolk=1.0/pow(rho,alpha);
	  tolk=max2(tolk,tol);	    
	  rho=rho*tao;
	}
      else if (all_cnorm<tolk && all_gnorm>tolk)
	{
	  tolk=1.0/pow(rho,alpha);
	  tolk=max2(tolk,tol);
	  rho=rho/tao;
	}
      else
	{
	  tolk=1.0/pow(rho,alpha);
	  tolk=max2(tolk,tol);
	}


      fprintf(fp1,"cnorm=%e gnorm=%e rho=%e",all_cnorm,all_gnorm,rho);

      lbfgsfloatval_t fx=0.0;
      for (i=0;i<npts;i++)
	{
	  for (j=i+1;j<npts;j++)
	    {
	      sumz=0.0;
	      for (d=0;d<dim;d++)
		{
		  sumz+=(z[i*dim+d]-z[j*dim+d])*(z[i*dim+d]-z[j*dim+d]);      
		}
	      fx+=1.0/sumz;
	    }
	}

      fprintf(fp1," %lf ",fx);  
      fprintf(fp1," %lf \n",fx/(npts*(npts-1)/2));  

      if (all_cnorm<=tol && all_gnorm<=tol)
	{
	  converg_flag=1;
	  break;
	}
    }
  tick_count parallel_t1 = tick_count::now();

  fprintf(fp1,"Finished %d \n",converg_flag);
  printf("Converge Flag= %d  Time Spent = %f ",converg_flag,(parallel_t1 - parallel_t0).seconds());
  fprintf(fp1," %f %e %e",(parallel_t1 - parallel_t0).seconds(),all_gnorm,all_cnorm);

  for (i=0;i<npts;i++)
    {
      sumz=0.0;
      for (d=0;d<dim;d++)
	{	  
	  fprintf(fp2,"%lf  ",z[i*dim+d]);
	  sumz+=z[i*dim+d]*z[i*dim+d];
	}
      fprintf(fp2," %lf \n",sumz);
    }

  lbfgsfloatval_t fx=0.0;
  for (i=0;i<npts;i++)
    {
      for (j=i+1;j<npts;j++)
	{
	  sumz=0.0;
	  for (d=0;d<dim;d++)
	    {
	      sumz+=(z[i*dim+d]-z[j*dim+d])*(z[i*dim+d]-z[j*dim+d]);      
	    }
	  fx+=1.0/sumz;
	}
    }

  fprintf(fp1," %lf ",fx);  
  printf(" fx=%lf ",fx);

  fprintf(fp1," %lf \n",fx/(npts*(npts-1)/2));  
  printf(" mean_fx=%lf \n",fx/(npts*(npts-1)/2));

  
  fclose(fp1);
  fclose(fp2);

  delete []x;
  delete []y;
  delete []z;
  delete []initz;

  return 0;
}
