#include <lbfgs.h>
#include <math.h>
#ifdef  _MSC_VER
#define inline  __inline
#endif/*_MSC_VER*/
#include <arithmetic_ansi.h>
#include <alm.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/timeb.h>
#include <string.h>

struct graph_instance{
    int npts;
    int dim;
    lbfgsfloatval_t *L;
    FILE* res_file;
    lbfgsfloatval_t gnorm;
    lbfgsfloatval_t cnorm;
    lbfgsfloatval_t xnorm;
    lbfgsfloatval_t avg_fx;
};

static lbfgsfloatval_t graph_function(
    void *instance,
    const lbfgsfloatval_t *x,
    lbfgsfloatval_t *g,
    const int n,
    const lbfgsfloatval_t step
    )
{
    struct graph_instance* t_ins;
    int i=0,j=0,index=0,m,dim,npts,k=0;
    lbfgsfloatval_t tmpdot;
    lbfgsfloatval_t fx; 
    t_ins = (struct graph_instance*) instance;
    dim = t_ins->dim;
    npts = t_ins->npts;
    fx = 0;
    vecset(g,0,n);

    for (i=0; i<npts; i++)
        for (j=0; j<npts; j++)
        {
            vecdot(&tmpdot, &x[i*dim], &x[j*dim], dim);
            fx += t_ins->L[i*npts+j]*tmpdot;
        }
    
    for (i=0; i<npts; i++)
    {
        for (j=0; j<npts; j++)
            for (k=0; k<dim; k++)
            {
                g[i*dim+k]+=t_ins->L[i*npts+j]*x[j*dim+k];
                g[j*dim+k]+=t_ins->L[i*npts+j]*x[i*dim+k];
            }
    }
    return fx;
}

static void sphere_constraint(
    void *instance,
    const lbfgsfloatval_t *x,
    lbfgsfloatval_t *cx,
    lbfgsfloatval_t *cxg,
    const int n,
    const int m,
    const int bool_gradient
    )
{
    struct graph_instance* t_ins;
    int dim,npts,i,j,k;
    lbfgsfloatval_t norm_x;

    t_ins = (struct graph_instance*) instance;
    npts = t_ins->npts;
    dim = t_ins->dim;
    vecset(cx, 0, m);
    
    if (bool_gradient)
        vecset(cxg, 0, n*m);

    for (k=0; k<dim; k++)
    {
        for (j=0; j<npts; j++)
            cx[k]+=x[j*dim+k];
    }


    for (i=0; i<npts; i++)
    {
        vecdot(&norm_x, x+i*dim, x+i*dim, dim);
        cx[i+dim] = norm_x - 1;
    }

    if (bool_gradient)
    {
        for (i=0; i<npts; i++)
        {
            for (k=0; k<dim; k++)
            {
                j=i*dim+k;
                cxg[j*m+k]=1;
                cxg[j*m+dim+i]=2*x[j];
            }
        }
    }

}

static int progress_lbfgs(
    void *instance,
    const lbfgsfloatval_t *x,
    const lbfgsfloatval_t *g,
    const lbfgsfloatval_t fx,
    const lbfgsfloatval_t xnorm,
    const lbfgsfloatval_t gnorm,
    const lbfgsfloatval_t step,
    int n,
    int k,
    int ls
    )
{
/*    int i=0;

    alm_instance_t *t_ins= (alm_instance_t*) instance;
    printf("Iteration %d:\n", k);
    printf("  fx = %f, x[0] = %f, x[1] = %f\n", fx, x[0], x[1]);
    printf("  xnorm = %f, gnorm = %f, step = %f, cxnorm = %f\n", xnorm, gnorm, step, t_ins->cx_norm);
    printf("  y = ");
    for (i=0; i<t_ins->m; i++)
    {
        printf("%f\t", t_ins->y[i]);
    }
    printf("\n");
    printf("\n"); */
    return 0;
}

static int progress_alm(
    void *instance,
    const lbfgsfloatval_t *x,
    const lbfgsfloatval_t *g,
    const lbfgsfloatval_t fx,
    const lbfgsfloatval_t xnorm,
    const lbfgsfloatval_t gnorm,
    const lbfgsfloatval_t cnorm,
    const lbfgsfloatval_t laxgnorm,
    const lbfgsfloatval_t fdiff,
    int n,
    int m,
    int k,
    int ls
    )
{
    int i=0;
    struct graph_instance* t_ins = (struct graph_instance*) instance;
    FILE *res_file= t_ins->res_file;
    fprintf(res_file, "%d %f %f %f %f\n", k, gnorm, cnorm, fx, 2*fx/m/(m-1));
    t_ins->gnorm = gnorm;
    t_ins->cnorm = cnorm;
    t_ins->xnorm = xnorm;
    t_ins->avg_fx = 2*fx/m/(m-1);
    return 0;
}

int main (int argc, char** argv)
{

    int npts = 3;
    int dim = 3;

    int n;
    int m;
    lbfgsfloatval_t *x,fx,xn;
    alm_parameter_t param;
    int ret;
    struct graph_instance * t_ins;
    int i=0,j=0;
    lbfgsfloatval_t sumx = 0;
    struct timeb start_time, end_time;
    char filepath[40]="./";
    char res_filename[50]="";
    char x_filename[50]="";
    char tmp[80]="";
    FILE *res_file=0, *x_file=0, *l_file=0;

    if (argc >= 2)
    {
        l_file=fopen(argv[1],"r");
        if (!l_file)
        {
            printf("Cannot open the matrix L file!\n");
            exit(1);
        }
    }
    else
    {
        printf("Please specify your file for matrix L as the first parameter\n.");
        exit(1);
    }
    if (argc >= 3)
        npts = atoi(argv[2]);
    if (argc >= 4)
        dim = atoi(argv[3]);
    if (argc >= 5)
        strcpy(filepath, argv[4]);

    sprintf(tmp, "result_alm_%d_%d.txt", npts, dim);
    strcat(res_filename, filepath);
    strcat(res_filename, tmp);
    tmp[0]='\0';
    sprintf(tmp, "x_alm_%d_%d.txt", npts, dim);
    strcat(x_filename, filepath);
    strcat(x_filename, tmp);
    res_file = fopen(res_filename, "w");
    x_file = fopen(x_filename, "w");

    if (!res_file || !x_file) {printf("Cannot open output files for writing results\n"); exit(1);}

    n = npts * dim;
    m = dim+npts;
    x = (lbfgsfloatval_t*) vecalloc(n*sizeof(lbfgsfloatval_t));

    srand(1);
    for (i=0; i<npts; i++)
    {
        sumx=0;
        for (j=0; j<dim; j++)
        {
            x[i*dim+j]=rand()%1000-500;
            sumx+=x[i*dim+j]*x[i*dim+j];
        }
        sumx = sqrt(sumx);
        for (j=0; j<dim; j++)
        {
            x[i*dim+j]/=sumx;
        }
    }
    
    t_ins = (struct graph_instance*) malloc(sizeof(struct graph_instance));
    t_ins->dim = dim;
    t_ins->npts = npts;
    t_ins->L = (lbfgsfloatval_t *) vecalloc(npts * npts *sizeof(lbfgsfloatval_t));
    for (i=0; i<npts; i++)
    {
        for (j=0; j<npts; j++)
        {
            fscanf(l_file, "%lf", &t_ins->L[i*npts+j]);
        }
    }
    t_ins->res_file = res_file;
    alm_parameter_init(&param);
    param.max_iter = 100;
    param.rho_times = 10;
    param.rho = 10;

    ftime(&start_time);
    ret=alm(n,m,x,&fx,graph_function, sphere_constraint, progress_lbfgs, progress_alm, t_ins, &param);
    ftime(&end_time);

    for (i=0; i<npts; i++)
    {
        for (j=0; j<dim; j++)
        {
            fprintf(x_file, "%f ",x[i*dim+j]); 
        }
        vec2norm(&xn,&x[i*dim],dim);
        fprintf(x_file,"%f", xn);
        fprintf(x_file, "\n");
    }
    fprintf(res_file, "Finish with converge flag %d\n", ret);
    printf("Finish with converge flag %d\n", ret);
    printf("Final Graph function value: %f \n", fx);
    printf("Time consumed: %f s.\n", (double)(end_time.time*1000+end_time.millitm-start_time.time*1000+end_time.millitm)/1000);
    fprintf(res_file, "%f %f %f %f %f\n", (double)(end_time.time*1000+end_time.millitm-start_time.time*1000+end_time.millitm)/1000, t_ins->gnorm, t_ins->cnorm, fx, t_ins->avg_fx);

    fclose(l_file);
    fclose(res_file);
    fclose(x_file);
    vecfree(x);
    vecfree(t_ins->L);
    free(t_ins);
}
